DATE_TIME = "Date Time"
TIMESTAMP = 'Timestamp'
TICK_NUMBER = "Tick Number"
OPEN_PRICE = "Open"
HIGH_PRICE = "High"
LOW_PRICE = "Low"
CLOSE_PRICE = "Close"

CUMULATIVE_TICKS = "Cumulative Ticks"
CUMULATIVE_DOLLAR = "Cumulative Dollar Value"
THRESHOLD = "Threshold"

CUMULATIVE_VOLUME = "Cumulative Volume"
CUMULATIVE_BUY_VOLUME = "Cumulative Buy Volume"
CUMULATIVE_SELL_VOLUME = "Cumulative Sell Volume"

CUMULATIVE_θ = "Cumulative θ"
CUMULATIVE_BUY_θ = "Cumulative Buy θ"
CUMULATIVE_SELL_θ = "Cumulative Sell θ"

EXPECTED_IMBALANCE = "expected_imbalance"
EXPECTED_TICKS_NUMBER = "exp_num_ticks"

EXPECTED_BUY_IMBALANCE = "exp_imbalance_buy"
EXPECTED_SELL_IMBALANCE = "exp_imbalance_sell"
EXPECTED_BUY_TICKS_PROPORTION = "exp_buy_ticks_proportion"
BUY_TICKS_NUMBER = "buy_ticks_num"

N_TICKS_ON_BAR_FORMATION = "Number of ticks while bar is formed."

PREVIOUS_TICK_RULE = "Previous tick rule"
EXPECTED_IMBALANCE_WINDOW = "Expected Imbalance Window"

PREVIOUS_BARS_N_TICKS_LIST = "List of previous bars number of ticks"
PREVIOUS_TICK_IMBALANCES_LIST = "List of previous tick imbalances"

PREVIOUS_TICK_IMBALANCES_BUY_LIST = "List of previous (buy) tick imbalances"
PREVIOUS_TICK_IMBALANCES_SELL_LIST = "List of previous (sell) tick imbalances"

PREVIOUS_BARS_BUY_TICKS_PROPORTIONS_LIST = "List of previous bars buy ticks proportion"

N_PREVIOUS_BARS_FOR_EXPECTED_N_TICKS_ESTIMATION = "Window size for E[T]"

